#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"

#include "resource.h"

#include "Delay.h"
#include "LowpassFilter.h"

#define TIMER_UPDATE 1

#pragma region  
double db2amp(double level)
{
	return pow(10.0, level / 20.0);
}

double amp2db(double amplitude)
{
	return 20.0 * log10(amplitude);
}
#pragma endregion

struct analog_subwoofer_params
{

	double freq;
	int order;

	double balance_level_lfe;
	double delay_lfe;


	// The constructor sets the default values.
	analog_subwoofer_params(	
		double freq=200, 
		int order=4, 
		double delay_lfe=0, 
		double balance_level_lfe=0
		) 
	{
		this->freq = freq;
		this->order = order;

		this->delay_lfe = delay_lfe;
		this->balance_level_lfe = balance_level_lfe;
	}

	void import(const analog_subwoofer_params& params )
	{
		memcpy(this,&params,sizeof(analog_subwoofer_params));
	}

	// GUID     
	static const GUID &g_get_guid() 
	{
		// {A633376F-5ECA-44F8-B5F4-E9CF44525924}
		static const GUID guid = { 0xa633376f, 0x5eca, 0x44f8, { 0xb5, 0xf4, 0xe9, 0xcf, 0x44, 0x52, 0x59, 0x24 } };
		return guid;
	}

	// Read data from a preset.
	bool set_data(const dsp_preset& p_dsp_preset) 
	{
		if (p_dsp_preset.get_owner() != g_get_guid()) return false;		
		if (p_dsp_preset.get_data_size() != sizeof(analog_subwoofer_params)) return false;

		memcpy(this,p_dsp_preset.get_data(),sizeof(analog_subwoofer_params));
		return true;
	}

	// Write data to a preset.
	void get_data(dsp_preset& p_dsp_preset) 
	{
		p_dsp_preset.set_owner(g_get_guid());		
		p_dsp_preset.set_data(this, sizeof(analog_subwoofer_params));
	}
};

// Our configuration dialog is implemented as a modal dialog.
// The helper class implements some safety measures to avoid
// opening multiple non-nested modal dialogs.
class dialog_analog_subwoofer : public dialog_helper::dialog_modal
{
private:
	LARGE_INTEGER meas_frequency, meas_last_update, meas_current;		
	bool is_dirty;
	bool is_dirty_range;
	const dsp_preset& m_old_data;
	analog_subwoofer_params params;
	dsp_preset_edit_callback& m_callback;

public:
	dialog_analog_subwoofer(const dsp_preset& p_dsp_preset, dsp_preset_edit_callback& p_callback) : m_old_data(p_dsp_preset), m_callback(p_callback), is_dirty(false) 
	{
		is_dirty_range = false;
		params.set_data(m_old_data);
	}

	virtual BOOL on_message(UINT msg, WPARAM wp, LPARAM lp) 
	{
		double freq=0, db=0;
		RECT r;
		POINT pt;

		switch(msg) 
		{
		case WM_INITDIALOG:
			{
				QueryPerformanceFrequency(&meas_frequency);
				SetTimer( get_wnd(), TIMER_UPDATE, 20, NULL );

				//
				HWND slider = GetDlgItem(get_wnd(), IDC_SLIDER_FREQ);
				SendMessage(slider, TBM_SETRANGE, TRUE, MAKELONG(50, 500));
				SendMessage(slider, TBM_SETPOS, TRUE, int(params.freq));
				SendMessage(slider, TBM_SETPAGESIZE, 0, 10);

				slider = GetDlgItem(get_wnd(), IDC_SLIDER_ORDER);
				SendMessage(slider, TBM_SETRANGE, TRUE, MAKELONG(0, 10));
				SendMessage(slider, TBM_SETPOS, TRUE, params.order);
				SendMessage(slider, TBM_SETPAGESIZE, 0, 10);

				slider = GetDlgItem(get_wnd(), IDC_SLIDER_BALANCE_LFE);
				SendMessage(slider, TBM_SETRANGE, TRUE, MAKELONG(-120, 120));
				SendMessage(slider, TBM_SETPOS, TRUE, int(params.balance_level_lfe*10));
				SendMessage(slider, TBM_SETTIC, 0, 0);
				SendMessage(slider, TBM_SETTIC, 0, -60);
				SendMessage(slider, TBM_SETTIC, 0, 60);
				SendMessage(slider, TBM_SETPAGESIZE, 0, 10);

				slider = GetDlgItem(get_wnd(), IDC_SLIDER_DELAY_LFE);
				SendMessage(slider, TBM_SETRANGE, TRUE, MAKELONG(-200, 200));
				SendMessage(slider, TBM_SETPOS, TRUE, int(params.delay_lfe*10));
				SendMessage(slider, TBM_SETTIC, 0, 0);
				SendMessage(slider, TBM_SETTIC, 0, -100);
				SendMessage(slider, TBM_SETTIC, 0, 100);
				SendMessage(slider, TBM_SETPAGESIZE, 0, 10);

				//
				update_display();
			}
			break;

		case WM_HSCROLL:
		case WM_VSCROLL:
			{
				is_dirty = true;
				update_display();
			}
			break;

			//----------------------
		case WM_TIMER:
			switch(wp)
			{
			case TIMER_UPDATE:
				QueryPerformanceCounter(&meas_current);
				LONGLONG period = (meas_current.QuadPart - meas_last_update.QuadPart) * 1000 / meas_frequency.QuadPart;
				if (period>100 && is_dirty) 
				{
					dsp_preset_impl data;
					params.get_data(data);
					m_callback.on_preset_changed(data);
					is_dirty = false;
					is_dirty_range = false;
				}		
				break;
			}
			break;

		case WM_DESTROY:
			KillTimer( get_wnd(), TIMER_UPDATE);
			break;
			//----------------------
		case WM_COMMAND:
			switch (LOWORD(wp)) 
			{
			case IDOK:
				if(is_dirty_range)
				{
					dsp_preset_impl data;
					params.get_data(data);
					m_callback.on_preset_changed(data);
					is_dirty_range = false;
				}
				end_dialog(1);				
				break;

			case IDCANCEL:
				m_callback.on_preset_changed(m_old_data);
				end_dialog(0);
				break;
			}
			break;
		}
		return 0;
	}

private:
	void update_display() 
	{
		// freq
		params.freq = SendDlgItemMessage(get_wnd(), IDC_SLIDER_FREQ, TBM_GETPOS, 0, 0);
		uSetDlgItemText(get_wnd(), IDC_STATIC_FREQ_VALUE,
			pfc::string_formatter() << pfc::format_int((int)params.freq) << " Hz");

		// order
		params.order = SendDlgItemMessage(get_wnd(), IDC_SLIDER_ORDER, TBM_GETPOS, 0, 0);
		uSetDlgItemText(get_wnd(), IDC_STATIC_ORDER_VALUE,
			pfc::string_formatter() << params.order);

		// delay lfe
		params.delay_lfe = 0.1* SendDlgItemMessage(get_wnd(), IDC_SLIDER_DELAY_LFE, TBM_GETPOS, 0, 0);
		uSetDlgItemText(get_wnd(), IDC_STATIC_DELAY_LFE_VALUE,
			pfc::string_formatter() << pfc::format_float((float)params.delay_lfe,0,1) << " ms");

		// balance lfe
		params.balance_level_lfe = 0.1* SendDlgItemMessage(get_wnd(), IDC_SLIDER_BALANCE_LFE, TBM_GETPOS, 0, 0);
		uSetDlgItemText(get_wnd(), IDC_STATIC_BALANCE_LFE_VALUE,
			pfc::string_formatter() << pfc::format_float((float)params.balance_level_lfe,0,1) << " dB");

		//
		QueryPerformanceCounter(&meas_last_update);
	}	
};

// Finally, this is our DSP class. We need a few extra methods
// to support presets.
class dsp_analog_subwoofer : public dsp_impl_base
{
	analog_subwoofer_params params;

	pfc::array_t<audio_sample> buffer;
	int last_srate;
	int last_channel_count;
	double latency;

	Delay* delay_left;
	Delay* delay_right;
	Delay* delay_center;
	Delay* delay_lfe;
	Delay* delay_backleft;
	Delay* delay_backright;

	LowpassFilter* lp;

	double amp_src;
	double amp_lfe;

public:
	dsp_analog_subwoofer(const dsp_preset & p_data) 
	{
		set_data(p_data);
		last_srate = 0;
		last_channel_count = 0;			

		delay_left = new Delay(16000);
		delay_right = new Delay(16000);
		delay_center = new Delay(16000);
		delay_lfe = new Delay(16000);
		delay_backleft = new Delay(16000);
		delay_backright = new Delay(16000);
		
		amp_src=1;
		amp_lfe=1;

		lp = new LowpassFilter(params.order);
	} 

	~dsp_analog_subwoofer()
	{
		delete delay_left;
		delete delay_right;
		delete delay_center;
		delete delay_lfe;
		delete delay_backleft;
		delete delay_backright;

		delete lp;
	}

	static GUID g_get_guid() 
	{
		return analog_subwoofer_params::g_get_guid();
	}

	static void g_get_name(pfc::string_base & p_out) 
	{
		p_out = "Analog Subwoofer";
	}

	// Return if we have a configuration popup (we do).
	static bool g_have_config_popup() 
	{
		return true;
	}

	// Show our configuration popup.
	// The provided callback is used to report configuration changes back to the caller.
	static void g_show_config_popup(const dsp_preset & p_dsp_preset, HWND p_parent, dsp_preset_edit_callback & p_callback) 
	{
		analog_subwoofer_params params;
		if (params.set_data(p_dsp_preset)) 
		{
			dialog_analog_subwoofer dlg(p_dsp_preset, p_callback);
			dlg.run(IDD_CONFIG, p_parent);
		}
	}

	// Return our default preset.
	static bool g_get_default_preset(dsp_preset & p_dsp_preset) 
	{
		analog_subwoofer_params().get_data(p_dsp_preset);
		return true;
	}

	// Read params from the provided preset.
	bool set_data(const dsp_preset & p_dsp_preset) 
	{		
		return params.set_data(p_dsp_preset);
	}

	// Process chunk.
	virtual bool on_chunk(audio_chunk * chunk, abort_callback & p_abort) 
	{	
		unsigned channel_count = chunk->get_channels();
		int srate = chunk->get_srate();
		if(last_srate == 0)
		{
			update_parameters(srate, channel_count);
		}
		else
		{
			if(last_srate != srate || last_channel_count != channel_count)
			{
				process_tail();
				update_parameters(srate, channel_count);
			}
		}
		last_srate = chunk->get_srate();		
		last_channel_count = channel_count;

		//
		t_size sample_count = chunk->get_sample_count();
		audio_sample* input = chunk->get_data();

		buffer.grow_size(sample_count*(max(channel_count,2)+2));

		audio_sample* output = buffer.get_ptr();

		//
		switch(channel_count)
		{
		case 1:
			process_data_mono(input, output, sample_count);
			break;
		case 2:
			process_data_stereo(input, output, sample_count);
			break;
		case 3:
			process_data_triple(input, output, sample_count);
			break;
		case 4:
			process_data_quadro(input, output, sample_count);
			break;
		case 5:
			process_data_penta(input, output, sample_count);
			break;
		default:
			return true;
		}
		return false;
	}

	virtual void on_endofplayback(abort_callback & p_abort) 
	{		
		if(latency==0) 
			return;
		process_tail();
	}

	void process_tail()
	{
	}

	//------------------------//
	// insert processed chunk //
	//------------------------//
	void process_data_mono(audio_sample* input, audio_sample* output, t_size sample_count)
	{
		audio_sample* store_output = output;
		t_size output_count = 0;

		for (t_size sample = 0; sample < sample_count; sample++) 
		{				
			double mono=0;
			if(input!=0)
			{
				mono = input[0];
				input+=1;
			}
			output_count++;

			// 
			double lfe = lp->process(mono);

			//   
			output[0] = amp_src * delay_left->process(mono);
			output[1] = amp_src * delay_right->process(mono);
			output[2] = amp_lfe * delay_lfe->process(lfe);		
			
			output+=3;
		}

		output = store_output;
		if(output_count>0)
		{			
			audio_chunk* newchunk = insert_chunk(output_count*3);		
			newchunk->set_data(output, output_count, 3, last_srate);
			newchunk->set_channels(3, newchunk->channel_front_left | newchunk->channel_front_right | newchunk->channel_lfe);
		}
	}

	void process_data_stereo(audio_sample* input, audio_sample* output, t_size sample_count)
	{
		audio_sample* store_output = output;
		t_size output_count = 0;

		for (t_size sample = 0; sample < sample_count; sample++) 
		{				
			double left=0;
			double right=0;
			if(input!=0)
			{
				left = input[0];
				right = input[1];
				input+=2;
			}
			output_count++;

			// 
			double mono =  0.5*(left + right);
			double lfe = lp->process(mono);

			//   
			output[0] = amp_src * delay_left->process(left);
			output[1] = amp_src * delay_right->process(right);
			output[2] = amp_lfe * delay_lfe->process(lfe);		
			
			output+=3;
		}

		output = store_output;
		if(output_count>0)
		{			
			audio_chunk* newchunk = insert_chunk(output_count*3);		
			newchunk->set_data(output, output_count, 3, last_srate);
			newchunk->set_channels(3, newchunk->channel_front_left | newchunk->channel_front_right | newchunk->channel_lfe);
		}
	}

	void process_data_triple(audio_sample* input, audio_sample* output, t_size sample_count)
	{
		audio_sample* store_output = output;
		t_size output_count = 0;

		for (t_size sample = 0; sample < sample_count; sample++) 
		{				
			double left=0;
			double right=0;
			double center=0;
			if(input!=0)
			{
				left = input[0];
				right = input[1];
				center = input[2];
				input+=3;
			}
			output_count++;

			// 
			double mono =  (left+right+center);
			double lfe = lp->process(mono);

			//   
			output[0] = amp_src * delay_left->process(left);
			output[1] = amp_src * delay_right->process(right);
			output[2] = amp_src * delay_center->process(center);
			output[3] = amp_lfe * delay_lfe->process(lfe);		
			
			output+=4;
		}

		output = store_output;
		if(output_count>0)
		{			
				audio_chunk* newchunk = insert_chunk(output_count*4);		
				newchunk->set_data(output, output_count, 4, last_srate);
				newchunk->set_channels(4, newchunk->channel_front_left | newchunk->channel_front_right | newchunk->channel_front_center | newchunk->channel_lfe);
		}
	}

	void process_data_quadro(audio_sample* input, audio_sample* output, t_size sample_count)
	{
		audio_sample* store_output = output;
		t_size output_count = 0;

		for (t_size sample = 0; sample < sample_count; sample++) 
		{				
			double left=0;
			double right=0;
			double backleft=0;
			double backright=0;
			if(input!=0)
			{
				left = input[0];
				right = input[1];
				backleft = input[2];
				backright = input[3];
				input+=4;
			}
			output_count++;

			// 
			double mono =  0.25*(left+right+backleft+backright);
			double lfe = lp->process(mono);

			//   
			output[0] = amp_src * delay_left->process(left);
			output[1] = amp_src * delay_right->process(right);
			output[2] = amp_src * delay_backleft->process(backleft);
			output[3] = amp_src * delay_backright->process(backright);
			output[4] = amp_lfe * delay_lfe->process(lfe);		
			
			output+=5;
		}

		output = store_output;
		if(output_count>0)
		{			
	audio_chunk* newchunk = insert_chunk(output_count*5);		
				newchunk->set_data(output, output_count, 5, last_srate);
				newchunk->set_channels(5, newchunk->channel_front_left | newchunk->channel_front_right | newchunk->channel_lfe | newchunk->channel_back_left | newchunk->channel_back_right);
	}
	}

	void process_data_penta(audio_sample* input, audio_sample* output, t_size sample_count)
	{
		audio_sample* store_output = output;
		t_size output_count = 0;

		for (t_size sample = 0; sample < sample_count; sample++) 
		{				
			double left=0;
			double right=0;
			double center=0;
			double backleft=0;
			double backright=0;
			if(input!=0)
			{
				left = input[0];
				right = input[1];
				center = input[2];
				backleft = input[3];
				backright = input[4];
				input+=5;
			}
			output_count++;

			// 
			double mono =  0.2*(left+right+center+backleft+backright);
			double lfe = lp->process(mono);

			//   
			output[0] = amp_src * delay_left->process(left);
			output[1] = amp_src * delay_right->process(right);
			output[2] = amp_src * delay_center->process(center);
			output[3] = amp_src * delay_backleft->process(backleft);
			output[4] = amp_src * delay_backright->process(backright);
			output[5] = amp_lfe * delay_lfe->process(lfe);		
			
			output+=6;
		}

		output = store_output;
		if(output_count>0)
		{			
				audio_chunk* newchunk = insert_chunk(output_count*6);		
			newchunk->set_data(output, output_count, 6, last_srate);
			newchunk->set_channels(6, newchunk->channel_front_left | newchunk->channel_front_right | newchunk->channel_front_center  | newchunk->channel_lfe | newchunk->channel_back_left | newchunk->channel_back_right);
		}
	}




	//#######################################################################

	void update_parameters(int srate, int channel_count)
	{
		// delays
		int dly_src = 0;
		int dly_lfe = floor(0.5+params.delay_lfe/1000.*srate);
		if(dly_lfe<0)
		{
			dly_src -= dly_lfe;
			dly_lfe = 0;
		}
		delay_left->SetDelay(dly_src);
		delay_right->SetDelay(dly_src);
		delay_lfe->SetDelay(dly_lfe);
		// amps
		double gain_src = 0;
		double gain_lfe = params.balance_level_lfe;
		if(gain_lfe>0)
		{
			gain_src -= gain_lfe;
			gain_lfe = 0;
		}
		amp_src = db2amp(gain_src);
		amp_lfe = db2amp(gain_lfe);
		// filter
		lp->SetFrequency(params.freq, srate);
	}

	virtual void on_endoftrack(abort_callback & p_abort) 
	{		
	}

	virtual void flush()
	{
		lp->reset();
	}

	virtual double get_latency() 
	{
		return latency;
	}

	virtual bool need_track_change_mark() 
	{
		return false;
	}
};

DECLARE_COMPONENT_VERSION(
"Analog Subwoofer",
	"1.0.0",
	"Subwoofer output with Butterworth filter.\n"
	"(c) 2017 Eugeniy Sokol, Magnitogorsk."
	);
static dsp_factory_t<dsp_analog_subwoofer> foo_dsp_analog_subwoofer;
